@@ -118,13 +118,14 @@
 //! hook up egui to Tetra's callbacks.
 
 #![warn(missing_docs)]
+#![allow(clippy::tabs_in_doc_comments)]
 
 pub use egui;
 
-use std::{fmt::Display, process::ExitStatus, sync::Arc, time::Instant};
+use std::{fmt::Display, sync::Arc, time::Instant};
 
 use copypasta::{ClipboardContext, ClipboardProvider};
-use egui::{epaint::ClippedShape, ClippedMesh, CtxRef, RawInput};
+use egui::{ClippedMesh, CtxRef, RawInput};
 use tetra::{
 	graphics::{self, BlendAlphaMode, BlendMode},
 	Event, TetraError,
@@ -295,46 +296,6 @@ fn egui_texture_to_tetra_texture(
 	)
 }
 
-/// An error that can occur when opening a URL or other path
-/// by clicking a hyperlink.
-#[derive(Debug)]
-pub enum OpenError {
-	/// An error occurred when interacting with the filesystem.
-	IoError(std::io::Error),
-	/// The program that opened the link finished unsuccessfully.
-	ProcessError(ExitStatus),
-}
-
-impl Display for OpenError {
-	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-		match self {
-			OpenError::IoError(error) => error.fmt(f),
-			OpenError::ProcessError(status) => status.fmt(f),
-		}
-	}
-}
-
-impl std::error::Error for OpenError {
-	fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
-		match self {
-			OpenError::IoError(error) => Some(error),
-			OpenError::ProcessError(_) => None,
-		}
-	}
-}
-
-impl From<std::io::Error> for OpenError {
-	fn from(error: std::io::Error) -> Self {
-		Self::IoError(error)
-	}
-}
-
-impl From<ExitStatus> for OpenError {
-	fn from(status: ExitStatus) -> Self {
-		Self::ProcessError(status)
-	}
-}
-
 /// An error that can occur when using egui-tetra.
 #[derive(Debug)]
 pub enum Error {
@@ -342,7 +303,7 @@ pub enum Error {
 	TetraError(TetraError),
 	/// An error occurred when opening a URL or other path
 	/// by clicking a hyperlink.
-	OpenError(OpenError),
+	OpenError(std::io::Error),
 	/// An error occurred when accessing the system's clipboard.
 	ClipboardError(Box<dyn std::error::Error + Send + Sync>),
 }
@@ -373,9 +334,9 @@ impl From<TetraError> for Error {
 	}
 }
 
-impl From<OpenError> for Error {
-	fn from(error: OpenError) -> Self {
-		Self::OpenError(error)
+impl From<std::io::Error> for Error {
+	fn from(v: std::io::Error) -> Self {
+		Self::OpenError(v)
 	}
 }
 
@@ -392,7 +353,7 @@ pub struct EguiWrapper {
 	ctx: CtxRef,
 	texture: Option<tetra::graphics::Texture>,
 	last_frame_time: Instant,
-	shapes: Option<Vec<ClippedShape>>,
+	meshes: Vec<(tetra::graphics::Rectangle<i32>, tetra::graphics::mesh::Mesh)>,
 }
 
 impl EguiWrapper {
@@ -403,7 +364,7 @@ impl EguiWrapper {
 			ctx: CtxRef::default(),
 			texture: None,
 			last_frame_time: Instant::now(),
-			shapes: None,
+			meshes: vec![],
 		}
 	}
 
@@ -521,6 +482,7 @@ impl EguiWrapper {
 		let now = Instant::now();
 		self.raw_input.predicted_dt = (now - self.last_frame_time).as_secs_f32();
 		self.last_frame_time = now;
+		self.meshes.clear();
 		self.ctx.begin_frame(self.raw_input.take());
 		if self.texture.is_none() {
 			self.texture = Some(egui_texture_to_tetra_texture(ctx, self.ctx.texture())?);
@@ -529,16 +491,20 @@ impl EguiWrapper {
 	}
 
 	/// Ends a GUI frame.
-	pub fn end_frame(&mut self) -> Result<(), Error> {
+	pub fn end_frame(&mut self, ctx: &mut tetra::Context) -> Result<(), Error> {
 		let (output, shapes) = self.ctx.end_frame();
-		self.shapes = Some(shapes);
+		if let Some(texture) = &self.texture {
+			let clipped_meshes = self.ctx.tessellate(shapes);
+			for ClippedMesh(rect, mesh) in clipped_meshes {
+				let rect = egui_rect_to_tetra_rectangle(rect);
+				let mesh = egui_mesh_to_tetra_mesh(ctx, mesh, texture.clone())?;
+				self.meshes.push((rect, mesh));
+			}
+		}
 
 		// open URLs that were clicked
 		if let Some(open_url) = &output.open_url {
-			let status = open::that(&open_url.url).map_err(|error| OpenError::IoError(error))?;
-			if !status.success() {
-				return Err(Error::OpenError(OpenError::ProcessError(status)));
-			}
+			open::that(&open_url.url)?;
 		}
 
 		// copy text to clipboard
@@ -553,19 +519,20 @@ impl EguiWrapper {
 	///
 	/// Note that this function changes the Tetra blend mode and
 	/// scissor state.
-	pub fn draw_frame(&mut self, ctx: &mut tetra::Context) -> Result<(), Error> {
-		if let (Some(texture), Some(shapes)) = (&self.texture, self.shapes.take()) {
-			graphics::set_blend_mode(ctx, BlendMode::Alpha(BlendAlphaMode::Premultiplied));
-			let clipped_meshes = self.ctx.tessellate(shapes);
-			for ClippedMesh(rect, mesh) in clipped_meshes {
-				graphics::set_scissor(ctx, egui_rect_to_tetra_rectangle(rect));
-				let mesh = egui_mesh_to_tetra_mesh(ctx, mesh, texture.clone())?;
-				mesh.draw(ctx, tetra::math::Vec2::zero());
-			}
-			graphics::reset_scissor(ctx);
-			graphics::reset_blend_mode(ctx);
+	pub fn draw_frame(&mut self, ctx: &mut tetra::Context) {
+		graphics::set_blend_mode(ctx, BlendMode::Alpha(BlendAlphaMode::Premultiplied));
+		for (rect, mesh) in &self.meshes {
+			graphics::set_scissor(ctx, *rect);
+			mesh.draw(ctx, tetra::math::Vec2::zero());
 		}
-		Ok(())
+		graphics::reset_scissor(ctx);
+		graphics::reset_blend_mode(ctx);
+	}
+}
+
+impl Default for EguiWrapper {
+	fn default() -> Self {
+		Self::new()
 	}
 }
 
@@ -579,9 +546,6 @@ impl EguiWrapper {
 #[allow(unused_variables)]
 pub trait State<E: From<Error> = Error> {
 	/// Called when it is time for the game to construct a GUI.
-	///
-	/// This is called at the beginning of the game loop, but
-	/// the GUI is not drawn until the end of the game loop.
 	fn ui(&mut self, ctx: &mut tetra::Context, egui_ctx: &egui::CtxRef) -> Result<(), E> {
 		Ok(())
 	}
@@ -634,11 +598,35 @@ impl<E: From<Error>> StateWrapper<E> {
 	}
 }
 
+/*
+A note about the order of events:
+
+Tetra's game loop is:
+- Poll and dispatch events
+- Run the update callback (may not happen every frame depending
+on the timestamp setting)
+- Run the draw callback
+
+We don't want to dispatch mouse and keyboard events to the
+gameplay code if egui wants them, but egui can't tell us if
+it wants them until after we call end_frame(). So we need
+to run the UI code at the beginning of the game loop (but
+not draw the result until the end of the game loop, of course).
+
+Here's the order of operations I settled on:
+- Whenever event is called, send it to the egui ctx and queue
+it up for later
+- At the beginning of update, run the UI callback and save the
+resulting meshes and scissor rectangles. Then, dispatch queued
+events to the gameplay code (unless the UI wanted them).
+- In the draw callback, draw gameplay first, then UI
+*/
+
 impl<E: From<Error>> tetra::State<E> for StateWrapper<E> {
 	fn update(&mut self, ctx: &mut tetra::Context) -> Result<(), E> {
 		self.egui.begin_frame(ctx)?;
 		self.state.ui(ctx, self.egui.ctx())?;
-		self.egui.end_frame()?;
+		self.egui.end_frame(ctx)?;
 
 		for event in self.events.drain(..) {
 			match &event {
@@ -672,7 +660,7 @@ impl<E: From<Error>> tetra::State<E> for StateWrapper<E> {
 
 	fn draw(&mut self, ctx: &mut tetra::Context) -> Result<(), E> {
 		self.state.draw(ctx, self.egui.ctx())?;
-		self.egui.draw_frame(ctx)?;
+		self.egui.draw_frame(ctx);
 		Ok(())
 	}
 
